#import "BFCast.h"
#import "BFParams.h"
#import <misc.h>
#import "Interval.h"

@implementation BFCast
/*"A BFCast is an object that holds all the forcecasting components of
  a bit forecast.  That means it has a BitVector (a thing called
  "conditions" in the code that keeps track of which world bits are
  being monitored) as well as other coefficients that are used to
  calculate forecasts. It has instance variables that record when the
  rule was last used, how many times it has been used, how accururate
  its predictions are, and so forth."*/

+createBegin: zone {
  BFCast *obj = [super createBegin: zone];

  obj->forecast = [Interval create];
  obj->lforecast = [Interval create];
  obj->variance = [Interval create];
  obj->strength = [Interval create];
  obj->a = [Interval create];
  obj->b = [Interval create];
  obj->c = [Interval create];
  obj->specfactor = [Interval create];
  obj->bitcost = [Interval create];

  return obj;
}

- createEnd
{
 if (!condwords || !condbits ){fprintf(stderr,"Must have condwords to create BFCast."); exit(1);}

  [forecast set: [Interval constant: 0.0]];
  count = 0;
  lastactive=1;
  specificity = 0;
  [variance set: [Interval constant: 999999999.0]];
  conditions= [ BitVector createBegin: [self getZone] ];
  [conditions setCondwords: condwords];
  [conditions setCondbits: condbits];
  conditions = [conditions createEnd];
  return self;
}

/*"The init is needed because BitVector has to be told how bit of a
  bit vector it will need"*/
+ init
{
  [BitVector init];
 return self;
}

/*"Free dynamically allocated memory"*/
- (void)drop
{
  [forecast drop];
  [lforecast drop];
  [variance drop];
  [strength drop];
  [a drop];
  [b drop];
  [c drop];
  [specfactor drop];
  [bitcost drop];
  [conditions drop];
  [super drop];
}

/*"Sets the number of words-worth's of conditions that are going to be used"*/       
- (void)setCondwords: (int)x
{
  condwords = x;
}

/*"Sets the number of bits. This is the number of aspect of the world that are monitored"*/
- (void)setCondbits: (int)x
{
  condbits = x;
}

/*"Null bits may be needed if the bit vector that is allocated is larger than the number of conditions being monitored. Since a bitvector allocates space in sizes of words, this might be important. Luckily, in the current design of ASM-2.0 (and after), there are no null bits."*/
- (void) setNNulls: (int) x
{
  nnulls = nnulls;
}

/*"Set the variable bitcost at x"*/
- (void) setBitcost: (Interval *) x
{
  [bitcost set: x];
  [x drop];
}
  
/*"Rather than individually set bits one by one, we might want to set
  all of them at once. That means we pass in a pointer to an array of
  words that is the "right size" for all the bits."*/
- (void) setConditions: (int *)x
{
  [conditions setConditions: x];
}

/*"Returns a pointer to an array, the integer representation of the
  conditions"*/
- (int *)getConditions
{
  return [conditions getConditions];
}

/*"Returns an object of type BitVector, the represnetation of
  conditions that is actually used inside this class or
  calculations"*/
- (BitVector *)getConditionsObject
{
  return conditions;
}


/*"For low level access to a full word's-worth of the condition"*/
- (void)setConditionsWord: (int)i To: (int)value
{
  [conditions setConditionsWord: i To: value];
}

/*"Returns the integer representation of the x'th word in the
  conditions"*/
- (int)getConditionsWord: (int)x
{
  return [conditions getConditionsWord: x];
}

/*"Sets the value of a bit in the conditions"*/
- (void)setConditionsbit: (int)bit To: (int)x
{
  [conditions setConditionsbit: bit To: x];
}

/*"If a bit is currently set to 0 ("don't care"), then change it to
  something else (1 or 2)"*/
- (void)setConditionsbit: (int)bit FromZeroTo: (int)x
{
  [conditions setConditionsbit: bit FromZeroTo: x];
}

/*"Returns 0,1,or 2, for a given bit in the conditions records"*/
- (int)getConditionsbit: (int)bit
{
  return [conditions getConditionsbit: bit];
}


- (void)maskConditionsbit: (int)bit
{
  [conditions maskConditionsbit: bit];
}


/*"Change a YES to a NO, and vice versa"*/
- (void)switchConditionsbit: (int)bit
{
  // conditions[WORD(bit)] ^= MASK[bit];
  [conditions switchConditionsbit: bit];
}

/*"Set a coefficient from demand equation"*/
- (void)setAval: (Interval *)x
{
  [a set: x];
  [x drop];
}

/*"Set b coefficient from demand equation"*/
- (void)setBval: (Interval *)x
{
  [b set: x];
  [x drop];
}

/*"Set c coefficient from demand equation"*/
- (void)setCval: (Interval *)x
{ 
  [c set: x];
  [x drop];
}

/*"Return a coefficient from demand equation"*/
- (Interval *)getAval
{
  return [[a clone] ret];
}

/*"Return b coefficient from demand equation"*/
- (Interval *)getBval
{
  return [[b clone] ret];
}

/*"Return c coefficient from demand equation"*/
- (Interval *)getCval
{
  return [[c clone] ret];
}

/*"Update the spec factor of this forecast. That means calculate:
specfactor= (condbits - nnulls - specificity)* bitcost
"*/
- (void)updateSpecfactor
{
  //was in BFagent: specfactor = 1.0/(1.0 + x*specificity);
  //but the bfagent.m way is so much nicer
  [specfactor set: [[
		     [[Interval castInt: condbits]
		       diff: [Interval castInt: nnulls]]
		      diff: [Interval castInt: specificity]]
		     prod: bitcost]
  ]; //follows bfagent.m
  
}

/*"Set the specfactor value of this forecast"*/
- (void)setSpecfactor: (Interval *)x
{
  [specfactor set: x];
  [x drop];
};

/*"Return the specfactor value of this forecast"*/
- (Interval *)getSpecfactor
{
  return [[specfactor clone] ret];
}

/*"Raise the specificity value of this forecast by one unit"*/
- (void)incrSpecificity
{
  ++ specificity;
}

/*"Reduce the specificity value of this forecast by one unit"*/
- (void)decrSpecificity
{
  --specificity;
}

/*"Set the specificity of this forecast"*/
- (void)setSpecificity: (int) x
{
  specificity = x;
}

/*"Return the specificity of this forecast"*/
- (int)getSpecificity
{
  return specificity;
}

/*"Set the variance of this forecast"*/
- (void)setVariance: (Interval *)x
{
  [variance set: x];
  [x drop];
}

/*"Return the variance of this forecast"*/
- (Interval *)getBFCastVariance
{
  return [[variance clone] ret];
}


/*"Set the time on which this forecast was last active to an input value"*/
- (void)setLastactive: (int)x
{
  lastactive = x;
}

/*"Return the time on which this forecast was last active"*/
- (int)getLastactive
{
  return lastactive;

}

/*"Return the value of count"*/
- (int)getCnt
{
  return count;
}


/*"Set the count variable to an inputted value"*/
- (void)setCnt: (int)x
{
  count = x;

}

/*"Increment this forecast's count variable"*/
- (int)incrCount;
{
  return ++count;
}

/*"Return strength of this forecast"*/
- (Interval *)getStrength
{
  return [[strength clone] ret];
}


/*"Set the strength value to an inputted value"*/
- (void) setStrength: (Interval *)x
{
  [strength set: x];
  [x drop];
}

/*"Set the previous forecast of this object to an inputted value"*/
- (void)setLforecast: (Interval *)x
{
  [lforecast set: x];
  [x drop];
}

/*"Get forecast from the previous time period"*/
- (Interval *)getLforecast
{
  return [[lforecast clone] ret];
}


/*"Set the forecast of this object to an inputted value"*/
- (void)  setForecast: (Interval *) x
{
  [forecast set: x];
  [x drop];
}

/*"Return forecast from this object"*/
- (Interval *)getForecast
{
  return [[forecast clone] ret];
}

/*"Calculate new forecast on basis of price and dividend information"*/
- (Interval *) updateForecastPrice: (Interval *)price Dividend: (Interval *)dividend
{
  [lforecast set: forecast];
  [forecast set: [[[a prod: [price sum: dividend]]
		    sum: [b prod: dividend]] sum: c]];
  [price drop];
  [dividend drop];
  return [[forecast clone] ret];
}


/*"Given an input forecast object, systematically ask it for
  all its IVARs and replace current settings with them."*/
- copyEverythingFrom: (BFCast *)from
{
 [forecast set: [from getForecast]];
 [lforecast set: [from getLforecast]];
 [variance set: [from getBFCastVariance]];
 [strength set:  [from getStrength]];
 [a set: [from getAval]];
 [b set: [from getBval]];
 [c set: [from getCval]];
 [specfactor set: [from getSpecfactor]];
 lastactive = [from getLastactive];
 specificity = [from getSpecificity];
 count = [from getCnt];
 [self setConditions: [from getConditions]];
 return self;
}

/*"Return some disgnostic information about the status of
  variables inside this forecast to the terminal"*/
- print
{
  int word;
  printf("BFCast print: forecast %s lforecast %s variance %s strength %s \n", [forecast printf: "%f"], [lforecast printf: "%f"], [variance printf: "%f"], [strength printf: "%f"]);
  for ( word=0; word < condwords; word++)
    [self printcond: word]; 
  printf("a %s b %s c %s specfactor %s, specificity %d count %d lastactive %d \n", [a printf: "%f"], [b printf: "%f"], [c printf: "%f"], [specfactor printf: "%f"], specificity,count,lastactive);

  [Interval printfclean];
  return self;
}

- printcond: (int)word
{
  [conditions printcond: word];
  return self;
}




@end










